//! Low-level Intermediate Representation of filters.
//!
//! Welcome to the machine room. Be careful to wear safety equipment.
//! The invariants in this module can be difficult to preserve.
//! `assert!` your way around here and watch your step.

use crate::filter::{self, Ast as Filter};
use crate::mir::{self, DefId, MirFilter};
use crate::path::{self, Path};
use alloc::{boxed::Box, vec::Vec};
use jaq_syn::filter::{AssignOp, BinaryOp, Fold, KeyVal};
use jaq_syn::{MathOp, Str};

#[derive(Debug, Clone, Default)]
struct View {
    /// indices of variables in the execution context that are currently visible
    vars: Vec<usize>,
    args: Vec<usize>,
    recs: Vec<usize>,
}

impl View {
    fn find_rec(&self, id: DefId, recs: &[Rec]) -> Option<usize> {
        self.recs.iter().find(|ridx| recs[**ridx].id == id).copied()
    }

    /// Keep only the variables and arguments bound in the given definition.
    fn truncate(&mut self, id: DefId, defs: &mir::Defs) {
        let (vars, args): (Vec<_>, Vec<_>) = defs.args(id).partition(|a| a.is_var());
        assert!(vars.len() <= self.vars.len());
        assert!(args.len() <= self.args.len());
        self.vars.truncate(vars.len());
        self.args.truncate(args.len());
    }
}

#[derive(Default)]
struct Ctx {
    /// number of variables in the execution context at the current point
    vars: usize,
    /// non-variable arguments, earliest bound first
    args: Vec<(MirFilter, DefId, View)>,
    /// list of recursively defined filters with their arity (only variable arguments)
    /// and the number of bound variables (including variable arguments)
    /// that must be in the context at the time of calling
    recs: Vec<Rec>,
}

pub struct Rec {
    id: DefId,
    vars_len: usize,
    filter: Filter,
}

pub fn root_def(defs: &mir::Defs) -> filter::Owned {
    //std::dbg!(defs);
    let root_id = 0;
    //let vars = defs.get(root_id).args.len();
    let mut ctx = Ctx::default();
    let view = View::default();
    let f = ctx.def(root_id, view, defs);
    let recs = ctx.recs.into_iter();
    let recs = recs.map(|rec| (defs.get(rec.id).arity(), rec.filter));
    filter::Owned::new(f, recs.collect())
}

// this has to be fulfilled for the IDs of the filters in any `view.recs`
fn sorted_and_unique<T: Ord + core::hash::Hash + Clone>(iter: impl Iterator<Item = T>) -> bool {
    use itertools::Itertools;
    let v: Vec<_> = iter.collect();
    let mut sorted = v.clone();
    sorted.sort();
    v.iter().all_unique() && v == sorted
}

impl Ctx {
    fn def(&mut self, id: DefId, mut view: View, defs: &mir::Defs) -> Filter {
        let def = defs.get(id);
        //std::dbg!("processing def", (&def.name, id, &view));

        let var_args = def.args.iter().filter(|a| a.is_var()).count();
        view.vars.extend(self.vars..self.vars + var_args);
        self.vars += var_args;

        let view_vars_len = view.vars.len();
        for rec_id in def.children.iter().filter(|cid| defs.get(**cid).recursive) {
            //std::dbg!("processing recursive child", rec_id);
            let rec = defs.get(*rec_id);
            assert!(rec.args.iter().all(|a| a.is_var()));
            let new_rec_idx = self.recs.len();
            view.recs.push(new_rec_idx);
            // put in a bogus filter that we replace later
            self.recs.push(Rec {
                id: *rec_id,
                vars_len: self.vars,
                filter: Filter::Id,
            });
            // std::dbg!(&self.recs);
            let f = self.def(*rec_id, view.clone(), defs);
            self.recs[new_rec_idx].filter = f;
        }
        assert_eq!(view.vars.len(), view_vars_len);

        //std::dbg!("rec defs done for", id, &view);
        //std::dbg!("now for the body of the definition", &def.body);

        let out = self.filter(def.body.clone(), id, view, defs);
        self.vars -= var_args;
        out
    }

    fn nonrec_call(
        &mut self,
        caller: DefId,
        callee: DefId,
        view: &View,
        nonvar_args: impl Iterator<Item = MirFilter>,
        defs: &mir::Defs,
    ) -> Filter {
        let last_common = defs.smallest_common_ancestor(caller, callee);
        let mut new_view = view.clone();
        new_view.truncate(last_common, defs);
        // we ban all recursive filters from the view that come "after"
        // the filter that we are calling
        // this relies on the property that filter IDs preserve order of definition
        new_view.recs.retain(|ridx| self.recs[*ridx].id < callee);
        //std::dbg!(&view, &new_view);

        for arg in nonvar_args {
            new_view.args.push(self.args.len());
            self.args.push((arg, caller, view.clone()));
        }

        self.def(callee, new_view, defs)
    }

    /// Convert a MIR filter contained in a definition `id` to a LIR filter.
    // TODO: operate on borrowed filter
    // the problem here is that for calls to arguments,
    // we need to access self.args, but we cannot pass the filter contained to it
    // to this function, because it mutably borrows self
    fn filter(&mut self, f: MirFilter, id: DefId, mut view: View, defs: &mir::Defs) -> Filter {
        let get = |f, ctx: &mut Self| Box::new(ctx.filter(f, id, view.clone(), defs));
        let of_str = |s: Str<_>, ctx: &mut Self| {
            let fmt = s.fmt.map_or(Filter::ToString, |fmt| *get(*fmt, ctx));
            use jaq_syn::string::Part;
            let mut iter = s.parts.into_iter().rev().map(|part| match part {
                Part::Str(s) => Filter::Str(s),
                Part::Fun(f) => Filter::Pipe(Box::new(*get(f, ctx)), false, Box::new(fmt.clone())),
            });
            let last = iter.next();
            iter.fold(last.unwrap_or_else(|| Filter::Str("".into())), |acc, x| {
                Filter::Math(Box::new(x), MathOp::Add, Box::new(acc))
            })
        };
        use mir::Filter as Expr;

        //std::dbg!(self.vars);
        //std::dbg!(&view.vars);
        //std::dbg!(&f.0);
        match f.0 {
            Expr::Var(v) => Filter::Var(self.vars - view.vars[v] - 1),
            Expr::Call(mir::Call::Native(n), args) => {
                let args = args.into_iter().map(|a| *get(a, self));
                Filter::Native(n, args.collect())
            }
            Expr::Call(mir::Call::Arg(a), args) => {
                //std::dbg!("arg call");
                assert!(args.is_empty());
                assert!(sorted_and_unique(view.args.iter()));
                // all accessible arguments are in the context
                assert!(view.args.iter().all(|aidx| *aidx < self.args.len()));
                // only actual arguments are accessible
                assert_eq!(
                    view.args.len(),
                    defs.args(id).filter(|a| !a.is_var()).count()
                );
                let (f, id, view) = self.args[view.args[a]].clone();
                self.filter(f, id, view, defs)
            }
            Expr::Call(mir::Call::Def(did), args) => {
                let args_len = self.args.len();
                let (var_arg_idxs, nonvar_arg_idxs) = defs.get(did).var_nonvar_arg_idxs();
                let var_args = var_arg_idxs.iter().map(|i| args[*i].clone());
                let nonvar_args = nonvar_arg_idxs.iter().map(|i| args[*i].clone());

                let var_args: Vec<_> = var_args
                    .map(|a| {
                        //std::dbg!(&view, self.vars);
                        let arg = self.filter(a, id, view.clone(), defs);
                        self.vars += 1;
                        // TODO: increase view.vars? probably not ...
                        arg
                    })
                    .collect();
                self.vars -= var_args.len();

                //std::dbg!(id);

                assert!(sorted_and_unique(
                    view.recs.iter().map(|ridx| self.recs[*ridx].id)
                ));
                // recursion!
                let out = if let Some(rec_idx) = view.find_rec(did, &self.recs) {
                    //std::dbg!("call a recursive filter!", did);
                    //  std::dbg!(&self.recs);
                    //  std::dbg!(&view.recs);
                    // arguments bound in the called filter and its ancestors
                    let vars_len = self.recs[rec_idx].vars_len;
                    //std::dbg!(var_args.len(), self.vars, vars_len);
                    Filter::Call {
                        id: rec_idx,
                        skip: self.vars - vars_len,
                    }
                } else {
                    self.nonrec_call(id, did, &view, nonvar_args, defs)
                };

                // here, we revert the order, because leftmost variable arguments are bound first, which means
                // they will appear *outermost* in the filter, thus have to be added *last* to the filter

                let filter = var_args.into_iter().rev().fold(out, |acc, arg| {
                    Filter::Pipe(Box::new(arg), true, Box::new(acc))
                });

                // std::dbg!("return from filter construction");
                self.args.truncate(self.args.len() - nonvar_arg_idxs.len());
                // we must be back to the original args len here
                assert_eq!(self.args.len(), args_len);

                filter
            }

            // variable-binding operators
            Expr::Binary(l, BinaryOp::Pipe(Some(_x)), r) => {
                let l = get(*l, self);

                view.vars.push(self.vars);
                self.vars += 1;
                let r = Box::new(self.filter(*r, id, view, defs));
                self.vars -= 1;

                Filter::Pipe(l, true, r)
            }
            Expr::Fold(typ, Fold { xs, init, f, .. }) => {
                let (xs, init) = (get(*xs, self), get(*init, self));

                view.vars.push(self.vars);
                self.vars += 1;
                let f = Box::new(self.filter(*f, id, view, defs));
                self.vars -= 1;

                Filter::Fold(typ, xs, init, f)
            }

            Expr::Id => Filter::Id,
            Expr::Num(mir::Num::Float(f)) => Filter::Float(f),
            Expr::Num(mir::Num::Int(i)) => Filter::Int(i),
            Expr::Str(s) => of_str(*s, self),
            Expr::Array(a) => {
                Filter::Array(a.map_or_else(|| Box::new(Filter::empty()), |a| get(*a, self)))
            }
            Expr::Object(o) => {
                let kvs = o.into_iter().map(|kv| match kv {
                    KeyVal::Filter(k, v) => (*get(k, self), *get(v, self)),
                    KeyVal::Str(k, v) => {
                        let k = of_str(k, self);
                        let v = match v {
                            None => Filter::Path(
                                Box::new(Filter::Id),
                                Path::from(path::Part::Index(k.clone())),
                            ),
                            Some(v) => *get(v, self),
                        };
                        (k, v)
                    }
                });
                Filter::Object(kvs.collect())
            }
            Expr::Try(f) => Filter::Try(get(*f, self), Box::new(Filter::empty())),
            Expr::Neg(f) => Filter::Neg(get(*f, self)),
            Expr::Recurse => Filter::recurse0(),

            Expr::Binary(l, BinaryOp::Pipe(None), r) => {
                Filter::Pipe(get(*l, self), false, get(*r, self))
            }
            Expr::Binary(l, BinaryOp::Comma, r) => Filter::Comma(get(*l, self), get(*r, self)),
            Expr::Binary(l, BinaryOp::Alt, r) => Filter::Alt(get(*l, self), get(*r, self)),
            Expr::Binary(l, BinaryOp::Or, r) => Filter::Logic(get(*l, self), true, get(*r, self)),
            Expr::Binary(l, BinaryOp::And, r) => Filter::Logic(get(*l, self), false, get(*r, self)),
            Expr::Binary(l, BinaryOp::Math(op), r) => {
                Filter::Math(get(*l, self), op, get(*r, self))
            }
            Expr::Binary(l, BinaryOp::Ord(op), r) => Filter::Ord(get(*l, self), op, get(*r, self)),
            Expr::Binary(l, BinaryOp::Assign(op), r) => {
                let (l, r) = (get(*l, self), get(*r, self));
                match op {
                    AssignOp::Assign => Filter::Assign(l, r),
                    AssignOp::Update => Filter::Update(l, r),
                    AssignOp::UpdateWith(op) => Filter::UpdateMath(l, op, r),
                }
            }
            Expr::Ite(if_thens, else_) => {
                let else_ = else_.map_or(Filter::Id, |else_| *get(*else_, self));
                if_thens.into_iter().rev().fold(else_, |acc, (if_, then_)| {
                    Filter::Ite(get(if_, self), get(then_, self), Box::new(acc))
                })
            }
            Expr::TryCatch(try_, catch_) => Filter::Try(
                get(*try_, self),
                catch_.map_or_else(|| Box::new(Filter::empty()), |c| get(*c, self)),
            ),
            Expr::Path(f, path) => {
                let f = get(*f, self);
                use jaq_syn::path::Part;
                let path = path.into_iter().map(|(p, opt)| match p {
                    Part::Index(i) => (path::Part::Index(*get(i, self)), opt),
                    Part::Range(lower, upper) => {
                        let lower = lower.map(|f| *get(f, self));
                        let upper = upper.map(|f| *get(f, self));
                        (path::Part::Range(lower, upper), opt)
                    }
                });
                Filter::Path(f, Path(path.collect()))
            }
        }
    }
}
